#! /usr/bin/python
import os
import string

# Author: Alexei Deinega, fdtd.kintechlab.com

# Usually to run a job on some supercomputer you need: 
# - make some job script where you specify executable file name, processors number, execution time, etc.,
# - submit this job script to a queue.
# Different supercomputers use different syntax for writing and submitting job scripts.
# It could be annoying to create job scripts and run them manually.
# To avoid this problem we made python script which allows you to submit jobs to a queue 
# at arbitrary supercomputer without specifying which supercomputer you are using at the moment.
# This is especially useful if you run your jobs at different supercomputers.
# To find which supercomputer you are using at the moment, our scripts use unix comand 'hostname'.

# You should use functions are qsub and qsub_appl (see documentation to them below)

# Before starting to use them, you should put instructions how to make job scripts 
# and submit them on your supercomputer in function make_script.

# This is already done for some supercomputers:
# SciNet, supercomputers of Kurchatov Institute, JSCC RAS, MSU, IHED RAS, Kintech Lab

########################### qsub ########################################

# This function creates directory workdir/worksubdir,
# copies executable file prog there, makes a job script SCRIPT.sh to run this executable file 
# with arguments args at np processors for some time (in seconds).
# if np=0, then job is serial.

# worksubdir and args can be strings or lists of strings
# in the last case, function will create more than one worksubdirs for corresponding arguments.
# job script will be sumbitted from last directory

# optional parameters:
# ppn - number of processors used per each node (default value ppn=0 means that all processors at the node can be used)
# task - job name (it will be displayed in submitted job list)
# fname - file name or list of file names which should be copied to workdir/worksubdir
# act - 1 (to submit a job script to the queue) or 0 (not do it)
# skip - some name. if file with this name (not '') will be found in the worksubdir, 
#        executable will not be copied there and job will not be submitted

def qsub(workdir,worksubdir,prog,args,time,np,ppn=0,task='',fname='',act=1,skip=''):

	if(type(worksubdir)!=type(args)):
		print 'Worksubdirs and args are of different type'
		return
	if(type(worksubdir)==list and len(worksubdir)!=len(args)):
		print 'Unequal amount of worksubdirs and args'
		return

	if np>0 and type(worksubdir)==list:
		if len(worksubdir)>1:
			print 'You cannot have more then one worksubdir for parallel job'
			return
		else:
			worksubdir=worksubdir[0]
			args=args[0]

	curdir=read_output('pwd',0) # remember current directory

	is_dir=os.access(workdir,os.F_OK)
	if not is_dir:	
        	os.mkdir(workdir)

	if (type(worksubdir)==str):
		worksubdir=[worksubdir]
	else:
		absworksubdir=[] # absolute path for worksubdirs

	for wdi in range(len(worksubdir)):
		os.chdir(workdir) # go to workdir
		wd=worksubdir[wdi]
		is_dir=os.access(wd,os.F_OK)
		if not is_dir:
	        	os.mkdir(wd)
		os.chdir(wd) # go to workdir/worksubdir
		abswd=read_output('pwd',0) # absolute path to worksubdir
		os.chdir(curdir) # come back to current directory

		if (type(args)==str):
			absworksubdir='%s' % (abswd)
		else:
			absworksubdir.append('%s'%(abswd))


		if (skip!=''):
			is_dir=os.access(abswd+'/'+skip,os.F_OK)
			if is_dir:
				print 'there is %s file in %s' % (skip,worksubdir)
				return ''

		print 'program '+prog

		is_dir=os.access(prog,os.F_OK)
		if not is_dir:
			print 'there is no %s file' % prog
			return

		os.system("cp "+prog+' '+'\''+abswd+'/.'+'\'')
		if(type(fname)==str):
			if fname=='':
				fname=[]
			else:
				fname=[fname]
		for f in fname:
			is_dir=os.access(f,os.F_OK)
			if not is_dir:
				print 'there is no %s file' % f
				return
			os.system('cp '+f+' '+'\''+abswd+'/.'+'\'')
		
		if wdi==len(worksubdir)-1: # make script for last directory in the list
			prog_name=prog[prog.rfind('/')+1:]
			res=make_script(absworksubdir,'./%s'%(prog_name),args,time,np,ppn,task,'SCRIPT.sh')
			if res[0]!='':	
				f=open(abswd+'/SCRIPT.sh','w') # script
				f.write(res[0])
				f.close()
		
			f=open(abswd+'/QSUB.sh','w') # command to submit script to a queue
			f.write(res[1])
			f.close()

			os.system('chmod u+x %s' % ('\''+abswd+'/QSUB.sh'+'\'')) # make 'QSUB.sh' as executable file

			if act==1: # submit script
				print 'script is submitting'
				os.chdir(abswd) # go to worksubdir
				output=read_output(res[1],1) # submit script to a queue
				os.chdir(curdir) # come back to current directory
			else:
				print 'script is created'
				output=''

#		os.chdir('..') # go to workdir
		
#	os.chdir(curdir) # come back to current directory
	return output

###########################################################################

# Usually, you have different home directories (where you compile your projects) and work directories 
# (where you run executable files and store the results of calculation) at different supercomputers.
# 
# You can specify these directories for each supercomputer and user name in function getdirs
# (to find which supercomputer and user name you are using at the moment, 
# unix comand 'hostname' and 'whoami' are used).
# Then you can use function qsub_appl, where these directories will be used automatically.

########################## qsub_appl ######################################

# This function creates workdir/worksubdir in work directory,
# copies there executable file prog from home directory,
# makes a job script there to run this executable file
# with arguments args at np processors for some time (in seconds).
# if np=0, then job is serial.

# worksubdir and args can be strings or lists of strings
# in the last case, function will create more than one worksubdirs for corresponding arguments

# optional parameters:
# ppn - number of processors used per each node (default value ppn=0 means that all processors at the node can be used)
# task - name which will be displayed in your submitted job list
# fname - file name or list of file names from home directory which should be copied to workdir/worksubdir
# act - 1 (to submit a script to the queue) or 0 (not do it)
# skip - some name. if file with this name (not '') will be found in the worksubdir, 
#        executable will not be copied there and job will not be submitted.

# symbol % in string prog will be substitute to specified arch name 
# (used with multi architectural makefiles, see http://make.paulandlesley.org/multi-arch.html)

def qsub_appl(workdir,worksubdir,prog,args,time,np,ppn=0,task='',fname='',act=1,skip=''):
	dirs=getdirs()
	home=dirs[0] # home directory
	work=dirs[1] # work directory
	arch=dirs[2] # architecture name
	if home=='':
	    return
	workdir=work+'/'+workdir
	prog=home+'/'+prog
	prog=prog.replace('%',arch)

	if(type(fname)==str):
		if(fname==''):
			fname=[]
		else:
			fname=[fname]
	for fi in range(len(fname)):
		fname[fi]=home+'/'+fname[fi]

	output=qsub(workdir,worksubdir,prog,args,time,np,ppn,task,fname,act)

	if act==1:
#		fstat = home+'/scripts/'+'qsub_result.dat'
		fstat='qsub_result.dat'
		if(type(worksubdir)==list):
			wsd=worksubdir[0]
		else:
			wsd=worksubdir
		stat = '%s %s %s %s %s %s' % (output,read_output('date',0),wsd,time,np,task)
		com = 'echo \'%s\' >> %s' % (stat,fstat)
		os.system(com)


########################## examples ####################################

# example for transport code

def qsub_transport(workdir,worksubdir,prjdir,args,time,np,ppn=0,task='',fname='',act=1,skip=''):

	prog=prjdir+'/transport/experiments/unix/_%/transport.x'
	if(fname!=''):
	    fname=prjdir+'/transport/experiments/'+fname
	qsub_appl(workdir,worksubdir,prog,args,time,np,ppn,task,fname,act,skip)

# example for EMTL (Electromagnetic Template Library, fdtd.kintechlab.com) code

def qsub_fdtd(workdir,worksubdir,prjdir,args,time,np,ppn=0,task='',fname='',act=1,skip=''):

	prog=prjdir+'/newfdtd/applications/experiments/unix/_%_mpi/fdtdii.x'
	if(fname!=''):
	    fname=prjdir+'/newfdtd/applications/experiments/'+fname
	qsub_appl(workdir,worksubdir,prog,args,time,np,ppn,task,fname,act,skip)

###################### some auxiliary functions ############################

# return output of some unix comand
def read_output(str, show):
	str=str+' > tmp'
	os.system(str)
	output=open('tmp').read()[:-1]
	os.system('rm tmp')
	if show==1:
		print output
	return output

# return string of hours:minutes:seconds format for number of seconds
def stime(sec):
	s=sec%60
	m=(sec/60)%60
	h=sec/(60*60)
	return "%d:%02d:%02d" % (h,m,s)

###################### make_script ##################################

# creates list of two strings:
# first string is script file corresponding for current architecture,
# second line - commands which are used to launch this script.
# workdir - workdir where executable file is
# prog - name of executable file
# args - arguments which are used when you run executable file
# time - executable time (supported for some architectures only)
# np - total number of processors. if np=0, then job is serial
# ppn - number of processors used per each node (supported for some architectures only)
# task - task name, it can be displayed in your submitted job list
# fname - script file name

def make_script(workdir,prog,args,time,np,ppn,task,fname):

	hostname=read_output('hostname',0)
	user=read_output('whoami',0)

	if ppn==0:
		if hostname=="biolab1.mipt.ru":
			ppn=2
		elif hostname=="scinet03" or hostname=="scinet04" or hostname=="gpc-f101n084":
			ppn=8
		elif hostname=="oivt-master":
			ppn=8
	if ppn>0:
		nodes=np/ppn # number of nodes
		if np%ppn>0:
			nodes=nodes+1

#----------------------- SciNet -----------------------

	if hostname=="scinet03" or hostname=="scinet04" or hostname=="gpc-f101n084":

# instruction how to make script on SciNet:
# https://support.scinet.utoronto.ca/wiki/index.php/SciNet_User_Support_Library

		if time>172800:
			time=172800
		st=stime(time)

		if np==0:
    
			result="""#!/bin/sh

#MOAB/Torque submission script for multiple serial jobs on SciNet GPC
#PBS -l nodes=1:ib:m32g:ppn=8,walltime=%s
#PBS -N %s 
cd $PBS_O_WORKDIR
#EXECUTION COMMAND; ampersand off 8 jobs and wait\n""" % (st,task)

			if(type(workdir)==str):
				workdir=[workdir]
				args=[args]

			for i in range(len(workdir)):
				result=result+"""(cd \'%s\'; %s %s) &\n""" % (workdir[i],prog,args[i])
				i+1

			result=result+'wait'

		else:

			result="""#!/bin/sh

#MOAB/Torque submission script for SciNet GPC (ethernet)
#PBS -l nodes=%d:ib:m32g:ppn=8,walltime=%s
#PBS -N %s 
cd $PBS_O_WORKDIR
cd \'%s\'
mpirun -np %d %s %s >& logfile.log """ % (nodes,st,task,workdir,np,prog,args)

		qsub = "qsub %s" % (fname)

#----------------------- radon.theory.northwestern.edu -----------------------

# instruction how to make script on radon:
# http://theory.northwestern.edu/wiki/index.php/Getting_Started_with_the_Clusters

	elif hostname=="radon.theory.northwestern.edu":

		result = """#!/bin/sh
#$ -pe mpich %d
#$ -N %s
#$ -cwd
#$ -q all.q
#$ -S /bin/bash
export LD_LIBRARY_PATH=/opt/gridengine/lib/lx26-amd64:/opt/openmpi/lib
export PATH=.:/opt/openmpi/bin:/usr/lib64/qt-3.3/bin:/usr/local/bin:/bin:/usr/bin:/opt/ganglia/bin:/opt/ganglia/sbin:/usr/java/latest/bin:/opt/pdsh/bin:/opt/rocks/bin:/opt/gridengine/bin/lx26-amd64
cd '%s'
mpirun -np %d %s %s >& logfile.log""" % (np,task,workdir,np,prog,args)

		qsub = "qsub %s" % (fname)

#----------------------- Kurchatov Institute -----------------------

	elif hostname=="n452" or hostname=="n454":

		if time<=600:
			time_queue='4g-10m'
		elif time<=3600:
			time_queue='1g-1h'
		elif time<=6*3600:
			time_queue='1g-6h'
		elif time<=24*3600:
			time_queue='1g-1d'
		elif time<=3*24*3600:
			time_queue='1g-3d'
		else:
			time_queue='1g-1w'
    
		result="""#!/bin/sh
#SBATCH -n %d""" % (np)
		result=result+"""
#SBATCH -o %j.out
#SBATCH -e %j.err"""

		result=result+"""
#SBATCH -p mpi-%s

source /opt/cluster
willuse mpi intel-compilers
echo "CPU list: $SLURM_JOB_NODELIST"
mpisub %s %s | tee mytask.log."$SLURM_JOBID" """ % (time_queue,prog,args)

		qsub = "sbatch -J %s %s" % (task,fname)

#----------------------- mvs -----------------------

	elif hostname=="mvs100k.jscc.ru" or hostname=="node1.eth0.mvs6k.jscc.ru":
		result=""
		qsub="""
cd %s
mpirun -maxtime %s -np %s -stdout output -stderr err %s %s 
""" % (workdir,time,np,prog,args)

        		
#----------------------- skif -----------------------

	elif hostname=="t60-2.parallel.ru":		

		result=""
        	qsub="""#!/bin/sh
cd %s
mpirun -maxtime %s -np %s %s %s 
		""" % (workdir,time,np,prog,args)

#----------------------- oivt-master -----------------------

	elif hostname=="oivt-master":		

	        result="""#!/bin/sh		
		
###PBS script to run task 		
#PBS -S /bin/bash		
#PBS -N %s 		
#PBS -q middle		
#PBS -l nodes=%s:ppn=%s		
#PBS -l walltime=%s		
		
#pwd		
#echo $PBS_O_WORKDIR		
		
# mvapich		
#cd $PBS_O_WORKDIR		
cd %s
pwd		

mpirun -n %s %s %s > output		
""" % (task,nodes,ppn,time,workdir,np,prog,args)

#/share/mpiexec-0.84/mpiexec -comm ib -n %s %s %s		
#""" % (task,nodes,ppn,time,np,prog,args)

#/share/mpiexec-0.84/mpiexec -comm ib %s %s		
#""" % (task,nodes,ppn,time,prog,args)
		qsub = "qsub %s" % (fname)		

#----------------------- nwo5.ihed.ras.ru -----------------------

	elif hostname=="nwo5.ihed.ras.ru":		

        	result="""#!/bin/sh		
	
#PBS -N %s -q test -l nodes=%s:ppn=%s -l walltime=%s
export LAM_MPI_SESSION_PREFIX=/tmp 
lamboot
lamnodes
cd %s 
pwd
time mpirun -np %s %s %s
lamhalt
touch done

""" % (task,nodes,ppn,time,workdir,np,prog,args)

		qsub = "qsub %s" % (fname)		

#----------------------- kintech -----------------------

	elif hostname=='c1.intranet.kintech.ru':
	
		result="""#!/bin/csh -f

#$ -N %s
#$ -pe mpich2 %d 
#$ -v MPIR_HOME=/usr/local/mpich
#$ -P photonic

echo "Got $NSLOTS slots."

setenv LD_LIBRARY_PATH "/opt/n1ge6/lib/lx24-x86:/opt/intel_fc_80/lib:/opt/intel_cc_80/lib"

setenv WORKDIR %s
cd $WORKDIR 

$MPIR_HOME/bin/mpirun -np $NSLOTS -machinefile $TMPDIR/machines %s %s > $WORKDIR/output
""" % (task,np,workdir,prog,args)

		qsub = "qsub %s" % (fname)

#----------------------- kintech2 -----------------------

	elif hostname=='kick.cluster.kintech.ru':
	
		result="""#!/bin/csh -f

#$ -N %s
#$ -pe impi* %d
#$ -v MPIR_HOME=/opt/intel/impi/3.1/
#$ -P photonics

echo "Got $NSLOTS slots."

source /opt/intel/ictce/3.1/ictvars.csh
setenv LD_LIBRARY_PATH /opt/intel/cce/10.1.008/lib
setenv PATH "$TMPDIR":"$PATH"

setenv WORKDIR %s 
cd $WORKDIR

$MPIR_HOME/bin64/mpirun -f $TMPDIR/machines -r ssh -n $NSLOTS %s %s > $WORKDIR/output
""" % (task,np,workdir,prog,args)

		qsub = "qsub %s" % (fname)

#----------------------- biolab1.mipt.ru -----------------------

	elif hostname=="biolab1.mipt.ru":

		result="""#!/bin/sh

#PBS -l nodes=%s:ppn=%s
#P -l nodes=node03:ppn=2
#P -l nodes=node03:ppn=2+node04:ppn=2+node08:ppn=2+node10:ppn=2+node11:ppn=2+node16:ppn=2

cd %s
env > myenv
sed 's/node\(..\)/node\\1g/' $PBS_NODEFILE > machines
echo $PBS_NODEFILE
lamboot ./machines
mpirun -v C %s %s > output
lamhalt
""" % (nodes,ppn,workdir,prog,args)

		qsub = "qsub %s -N %s" % (fname,task)

#-----------------------unknown -----------------------

	else:
		print 'unknown host'
		return['','']

	return [result,qsub]		

######################## getdirs #################################

# return home directory name, work directory name and arch name

def getdirs():

	hostname=read_output('hostname',0)
	user=read_output('whoami',0)

	home=''
	work=''
	arch=''

	if hostname=='scinet03' or hostname=='scinet04' or hostname=='gpc-f101n084':
		work='/scratch/s/sjohn/'
		arch='scinet'
	elif hostname=='radon.theory.northwestern.edu':
		arch='radon'
	elif hostname=='n452' or hostname=='n454':
		arch='kiae'
	elif hostname=='node1.eth0.mvs6k.jscc.ru':
		home='/gpfs/NETHOME/oivt3/'
		arch='mvs6ki'
	elif hostname=='mvs100k.jscc.ru':
		home='/gpfs/NETHOME/oivt3/'
		arch='mvs50k'
	elif hostname=='t60-2.parallel.ru':		
		arch='skif'
	elif hostname=='oivt-master':		
		arch='ihed'	
	elif hostname=='nwo5.ihed.ras.ru':		
		arch='ihed'	
	elif hostname=='c1.intranet.kintech.ru':
		home='/work/'
		arch='kintech'
	elif hostname=='kick.cluster.kintech.ru':
		home='/work/'
		arch='kintech2'
	elif hostname=='biolab1.mipt.ru':
		arch='gnu'
	else:
		print 'unknown host'
		return ['','','']

	if home=='':
		home=read_output('echo $HOME > tmp',0)
	else:
		home=home+user

	if work=='':
		work=home+'/runs'
	else:
		work=work+user+'/runs'

	return [home,work,arch]
